from ..utils.mapper import mapFromTo


from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from tkinter import Tk, mainloop, Scale, HORIZONTAL
import matplotlib.pyplot as plt
import numpy as np


def hist_normalizer(img : np.ndarray) -> None:
    '''
        Open a gui that helps to stretch the histogram of an image
     
        Parameters:
        -----------
            - img : np.ndarray 
                a WxHxB image, with width W, height H and B bands
        
        Returns:
        --------
        Nothing, it will display and image

        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
            [
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
            [
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        hist_normalizer(img)
        ```

        Output:
        -------
        Nothing, it will display an image
    '''
    

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')

    # Update plot
    def plot(img):
        # Clear axis
        axes[0].cla()
        axes[1].cla()
        
        # Plot image and histogram
        axes[0].set_title('Image')
        img = mapFromTo(img, wmin.get(), wmax.get(), 0, 1)
        axes[0].imshow(img)
    
        axes[1].set_title('Histogram')
        axes[1].set_xlabel('Mapped Pixel Values')
        axes[1].set_ylabel('Frequency')
    
        for k in range(img.shape[-1]):
            axes[1].hist(img[:,:,k].flatten(), 200, label = 'Band-{}'.format(k))   
            
        axes[1].legend()
        chart_type.draw()
    
    # Scale values using the sliders
    def scalevalue(value, other = img):
        image = np.clip(img, wmin.get(), wmax.get())    
        plot(image)
    
    figure, axes = plt.subplots(nrows = 1, ncols = 2, figsize=(10,5), dpi=100)
    
    # GUI
    root = Tk()
    root.title('Histrogram Scaler')
    #root.geometry('1000x620')
    #root.resizable(False, False)
    
    chart_type = FigureCanvasTkAgg(figure, root)
    chart_type.get_tk_widget().pack()
    
    wmin = Scale(root, from_=np.min(img), to=np.max(img), length = 400, label='Minimum', orient=HORIZONTAL, command = scalevalue)
    wmin.set(np.min(img))
    wmin.pack()
    wmax = Scale(root, from_=np.min(img), to=np.max(img), length = 400, label='Maximum', orient=HORIZONTAL, command = scalevalue)
    wmax.set(np.max(img))
    wmax.pack()
    
    mainloop()



import matplotlib.pyplot as plt
import numpy as np

def plot(img : np.ndarray, hist : bool = False) -> None:
    '''
        Plot a satellite image and its histogram.

        Only gray-scale or RGB-like are accepted.

        Parameters:
        ----------
            - img : np.ndarray
                a WxHxB image, with width W, height H and B bands (B can be 3 or 1)
            - hist : bool
                if True display also the histogram (dafault : False)
        
        Returns:
        --------
        Nothing, it will display the image

        Usage:
        ------

        ```python
        import numpy as np

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ]
            ]  
        ) 

        # Adding fake axis
        img = img[:,:,None]

        plot(img, hist=True)

        ```

        Output:
        -------
        Nothing, it will display the image
    '''
    

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')

    if ((img.shape[-1] != 1) and (img.shape[-1] != 3)):
        raise Exception("Error: numer of channels admitted are 1 or 3")


    ncols = 1
    if hist: ncols = 2
    
    fig, axes = plt.subplots(nrows = 1, ncols = ncols, figsize = (4*ncols, 4))

    if hist:
        axes[0].imshow(img)
        axes[0].set_title('Image')
        for b in range(img.shape[-1]):
            axes[1].hist(img[:,:,b].flatten(), 200, label='Band {}'.format(b))
        axes[1].legend()
        axes[1].set_title('Histogram')
    else:
        axes.imshow(img)
        axes.set_title('Image')
        
    fig.tight_layout()
    plt.show()


def bands_plot(img : np.ndarray, hist : bool = False) -> None:
    '''
        Plot a satellite image bands and relative histograms.

        Parameters:
        ----------
            - img : np.ndarray
                a WxHxB image, with width W, height H and B bands
            - hist : bool
                if True display also the histograms (dafault : False)
        
        Returns:
        --------
        Nothing, it will display the image

        Usage:
        ------

        ```python
        import numpy as np

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
             ]
            ]  
        ) 

        # Adding fake axis
        img = np.moveaxis(img, 0, -1)

        bands_plot(img, hist=True)

        ```

        Output:
        -------
        Nothing, it will display the image
    '''
    
    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')
    
    ncols = img.shape[-1]
    nrows = 1
    if hist: nrows = 2
    
    fig, axes = plt.subplots(nrows = nrows, ncols = ncols, figsize = (4*ncols, 4*nrows))

    for c in range(ncols):
        
        if hist:
            axes[0,c].imshow(img[:,:,c])
            axes[0,c].set_title('Bands {}'.format(c))
                
            axes[1,c].hist(img[:,:,c].flatten(), 200)
            axes[1,c].set_title('Histogram of Band {}'.format(c))
        else:
            axes[c].imshow(img[:,:,c])
            axes[c].set_title('Band {}'.format(c))

    fig.tight_layout()
    plt.show()


from sklearn.cluster import KMeans
import numpy as np

def pixel_clustering(img : np.ndarray, n_clusters : int = 3) -> np.ndarray:
    '''
        Image pixel clustering using Support Vector machine
        
        Parameters:
        -----------
            - img : np.ndarray  
                a WxHxB image, with width W, height H and B bands (channel last)
            - n_clusters : int  
                number of cluster to be identified
        
        Returns:
        --------
            - clusters : np.ndarray  
                clustered pixels

        Usage:
        ------

        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
             ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        clusters = pixel_clustering(img, n_clusters = 2)
        ```
        Output:
        ------
        ```
        array([[0, 0, 0],
               [0, 0, 1],
               [1, 1, 1]], dtype=int32)
       ```

    '''

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')
    
    image_2D = img.reshape(img.shape[0]*img.shape[1], img.shape[2])
    kmeans = KMeans(n_clusters = n_clusters, random_state = 0).fit(image_2D)
    clusters = np.array(kmeans.labels_).reshape(img.shape[0], img.shape[1])
    
    return clusters

def conv_clustering(): pass


from mayavi import mlab
import numpy as np


def plot3d(img : np.ndarray, animate : bool = False) -> None:
    '''
        3D-Plot of a satellite image.

        Parameters:
        ----------
            - img : np.ndarray
                a WxHxB image, with width W, height H and B bands (B can be 3 or 1)
            - animate : bool
                activate animation mode (dafault : False)
        
        Returns:
        --------
        Nothing, it will display the image

        Usage:
        ------

        ```python
        import numpy as np

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [0.9, 0.8, 0.7],  
                [0.6, 0.5, 0.4],  
                [0.3, 0.2, 0.1]
             ]
            ]  
        ) 

        # Making channel last
        img = np.moveaxis(img, 0, -1)

        plot3d(img, animate=True)

        ```

        Output:
        -------
        Nothing, it will display the image

    '''

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')

    fig = mlab.figure(figure='', bgcolor=(1, 1, 1), fgcolor=(0, 0, 0), size=(600, 600))
    v = mlab.volume_slice(img, slice_index=0, plane_orientation='z_axes', figure=fig)  # depth slice

    if animate:
        @mlab.animate
        def anim():
            ct = 0
            while True:
                if ct >= img.shape[-1]:
                    ct = 0
                v.mlab_source.scalars = img[:, :, ct]
                ct += 1
                yield
        anim()
    mlab.show()


def cube_plot(img : np.ndarray, save_path : str, band_thinkness : int = 3, cmap : str = 'jet', show : bool = True) -> None:
    '''
        Cube plot of a satellite image.

        Parameters:
        ----------
            - img : np.ndarray
                a WxHxB image, with width W, height H and B bands (B can be 3 or 1)
            - save_path : str
                path where to save figure
            - band_thinkness : int
                thikness of each satellite band
            - cmap : str
                colormap for the plot
            - show : bool
                if True will show the plot, othewise will be only saved to a file
        
        Returns:
        --------
        Nothing, it will display the image

        Usage:
        ------

        ```python
        import numpy as np

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [0.9, 0.8, 0.7],  
                [0.6, 0.5, 0.4],  
                [0.3, 0.2, 0.1]
             ]
            ]  
        ) 

        # Making channel last
        img = np.moveaxis(img, 0, -1)

        cube_plot(img, 'img.png', band_thikness=10, cmap='jet', show=True)

        ```

        Output:
        -------
        Nothing, it will display the image

    '''
    
    if len(img.shape) != 3: raise Exception("Error: lenght of img shape must be 3")


    img = np.repeat(img, band_thinkness, axis=-1)

    fig = mlab.figure(figure='', bgcolor=(1,1,1), fgcolor=(0, 0, 0), size=(600, 600))

    scalars = img  # specifying the data array

    # Crossline slices
    mlab.volume_slice(scalars, slice_index=0, plane_orientation='x_axes', figure=fig, colormap=cmap)  # crossline slice
    mlab.volume_slice(scalars, slice_index=img.shape[0], plane_orientation='x_axes', figure=fig, colormap=cmap)  # crossline slice

    # Inline slices
    mlab.volume_slice(scalars, slice_index=0, plane_orientation='y_axes', figure=fig, colormap=cmap)  # inline slice
    mlab.volume_slice(scalars, slice_index=img.shape[1], plane_orientation='y_axes', figure=fig, colormap=cmap)  # inline slice

    # Depth slices
    mlab.volume_slice(scalars, slice_index=img.shape[-1], plane_orientation='z_axes', figure=fig, colormap=cmap)  # depth slice
    mlab.volume_slice(scalars, slice_index=0, plane_orientation='z_axes', figure=fig, colormap=cmap)  # depth slice


    mlab.draw()
    mlab.savefig(filename=save_path)
    
    if show: mlab.show()
    else: 
        mlab.clf()
        mlab.close(scene=None, all=True)


def stacked_cube_plot(imgs : list, save_path : str, cmaps : list, band_thinkness : int = 3, spacing : int = 10, show : bool = True) -> None:
    '''
        Cubes plot of a satellite image.

        Parameters:
        ----------
            - imgs : list 
                list of WxHxB image (ndarray), with width W, height H and B bands (B can be 3 or 1)
            - save_path : str
                path where to save figure
            - cmaps : list
                list of colormap for the plot
            - band_thinkness : int
                thikness of each satellite band
            - spacing : int
                add empty space between cubes
            - show : bool
                if True will show the plot, othewise will be only saved to a file
        
        Returns:
        --------
        Nothing, it will display the image

        Usage:
        ------

        ```python
        import numpy as np

        img_1         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [0.9, 0.8, 0.7],  
                [0.6, 0.5, 0.4],  
                [0.3, 0.2, 0.1]
             ]
            ]  
        )

        img_2         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [0.9, 0.8, 0.7],  
                [0.6, 0.5, 0.4],  
                [0.3, 0.2, 0.1]
             ]
            ]  
        ) 

        # Making channel last
        img_1 = np.moveaxis(img_1, 0, -1)
        img_2 = np.moveaxis(img_2, 0, -1)


        cube_plot([img_1, img_2], 'img.png', band_thikness=10, spacing=10, cmap='jet', show=True)

        ```

        Output:
        -------
        Nothing, it will display the image

    '''
    
    fig = mlab.figure(figure='', bgcolor=(1,1,1), fgcolor=(0, 0, 0), size=(600, 600))
    
    for i, img in enumerate(imgs):

        if len(img.shape) != 3: raise Exception("Error: lenght of img shape must be 3")


        img = np.repeat(img, band_thinkness, axis=-1)
        scalars = img  # specifying the data array

        # Crossline slices
        mlab.volume_slice(scalars, slice_index=0, plane_orientation='x_axes', figure=fig, colormap=cmaps[i])  # crossline slice
        mlab.volume_slice(scalars, slice_index=img.shape[0], plane_orientation='x_axes', figure=fig, colormap=cmaps[i])  # crossline slice

        # Inline slices
        mlab.volume_slice(scalars, slice_index=0, plane_orientation='y_axes', figure=fig, colormap=cmaps[i])  # inline slice
        mlab.volume_slice(scalars, slice_index=img.shape[1], plane_orientation='y_axes', figure=fig, colormap=cmaps[i])  # inline slice

        # Depth slices
        mlab.volume_slice(scalars, slice_index=img.shape[-1] + (i+1)*spacing, plane_orientation='z_axes', figure=fig, colormap=cmaps[i])  # depth slice
        mlab.volume_slice(scalars, slice_index=i*spacing, plane_orientation='z_axes', figure=fig, colormap=cmaps[i])  # depth slice


    mlab.draw()
    mlab.savefig(filename=save_path)
    
    if show: mlab.show()
    else: 
        mlab.clf()
        mlab.close(scene=None, all=True)



from ..utils.paths import get_path_gui

import matplotlib.pyplot as plt
import numpy as np
import rasterio
import netCDF4


def write(image : np.ndarray, path : str, meta : dict) -> None:
    '''
        Save an image and its metadata given a path.

        Supported data format

        RASTERIO_EXTENSIONS   = ['.tif', '.tiff', '.geotiff']  
        MATPLOTLIB_EXTENSIONS = ['.png', '.jpg', 'jpeg', 'jp2']
        NETCDF4_EXTENSIONS    = ['.nc']

        Data must always be in channel last format.

        If image extension is in MATPLOTLIB_EXTENSIONS, metadata can be None.
        If image extension is in NETCDF4_EXTENSIONS, metadata can be None.

        Parameters:
        -----------
            - image : np.ndarray
                the WxHxC image to be saved, with W width, H height and B bands (channel last)
            - path : str 
                position of the image, if None the function will ask for the image path using a menu
            - meta : dict
                metadata for the image to be saved
        
        Returns:
        --------
        Nothing, the image will be saved

        Usage:
        ------

        ```python
        import numpy as np

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
             ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        write(img, 'path/to/save/img.png')
        
        ```
        Output:
        -------
        Nothing, the image will be saved


    '''

    RASTERIO_EXTENSIONS   = ['.tif', '.tiff']
    MATPLOTLIB_EXTENSIONS = ['.png', '.jpg', 'jpeg', 'jp2']
    NETCDF4_EXTENSIONS    = ['.nc']


    if path is None:
        path = get_path_gui()

    if any(frmt in path for frmt in RASTERIO_EXTENSIONS):
        if image!=None:
            meta.update({'driver':'GTiff',
                            'width':image.shape[0],
                            'height':image.shape[1],
                            'count':image.shape[2],
                            'dtype':'float64'})

        with rasterio.open(fp=path, mode='w',**meta) as dst:
            for count in range(image.shape[2]):
                dst.write(image[:,:,count], count)

    elif any(frmt in path for frmt in MATPLOTLIB_EXTENSIONS):
        plt.imsave(path, image)
    elif any(frmt in path for frmt in NETCDF4_EXTENSIONS):
        raise Exception('Error: [under dev] currently netCDF4 files can not be saved!')
    else:
        raise Exception('Error: file can not be saved, format not supported!')



from ..utils.paths import get_path_gui

import matplotlib.pyplot as plt
import numpy as np
import rasterio
import netCDF4


def load(path : str) -> [np.ndarray or dict, dict, list]:
    '''
        Load an image and its metadata given its path.

        Supported data format

        RASTERIO_EXTENSIONS   = ['.tif', '.tiff', '.geotiff']  
        MATPLOTLIB_EXTENSIONS = ['.png', '.jpg', 'jpeg', 'jp2']
        NETCDF4_EXTENSIONS    = ['.nc']

        Returns always data in channel last format.

        If image extension is in MATPLOTLIB_EXTENSIONS, metadata and bouns will be None.
        If image extension is in NETCDF4_EXTENSIONS, metadata and bounds will be None.
        
        Parameters:
        -----------
            - path : str
                position of the image, if None the function will ask for the image path using a menu

        Returns:
        --------
            - data : np.ndarray or list
                WxHxB image, with W width, H height and B bands

            - metadata : dict
                dictionary containing image metadata
            
            - bounds : list
                list containing geo bounds
        
        Usage:
        ------
        ```python
            img = load(None)
        ``` 
        or
        ```python
            img = load("path/to/image.png")
        ``` 

        Output:
        -------
        ```
        (
            array([[[5872., 5532., 5516., ...,    0.,    0., 1024.],  
                    [5872., 5588., 5451., ...,    0.,    0., 1024.],  
                    [5872., 5606., 5333., ...,    0.,    0., 1024.],  
                    ...,  
                    [2672., 2602., 2368., ...,    0.,    0., 1024.],  
                    [2672., 2689., 2394., ...,    0.,    0., 1024.],  
                    [2672., 2705., 2431., ...,    0.,    0., 1024.]],  
                    ...,  
                    [[1571., 1318., 1167., ...,    0.,    0.,    0.],  
                    [1571., 1206., 1113., ...,    0.,    0.,    0.],  
                    [1571., 1230., 1094., ...,    0.,    0.,    0.],  
                    ...,  
                    [1330., 1044.,  837., ...,    0.,    0.,    0.],  
                    [1330., 1045.,  842., ...,    0.,    0.,    0.],  
                    [1330., 1032.,  833., ...,    0.,    0.,    0.]]]),  
            
            {'driver': 'GTiff', 'dtype': 'float64', 'nodata': None, 'width': 1043, 'height': 1040, 'count': 16, 'crs': CRS.from_epsg(32632), 'transform': Affine(10.0, 0.0, 638640.0,
       0.0, -10.0, 5084590.0), 'blockxsize': 256, 'blockysize': 256, 'tiled': True, 'compress': 'lzw', 'interleave': 'pixel'},  

       BoundingBox(left=638640.0, bottom=5074190.0, right=649070.0, top=5084590.0))  
        )

        ```
    '''
    
    
    RASTERIO_EXTENSIONS   = ['.tif', '.tiff', '.geotiff']
    MATPLOTLIB_EXTENSIONS = ['.png', '.jpg', 'jpeg', 'jp2']
    NETCDF4_EXTENSIONS    = ['.nc']
    NUMPY_EXTENSIONS      = ['.npy', '.npz']
    
    
    if path is None:
        path = get_path_gui()
    
    if any(frmt in path for frmt in RASTERIO_EXTENSIONS):
        with rasterio.open(path) as src:
            data = src.read()
            metadata = src.profile
            bounds = src.bounds
        data = np.moveaxis(data, 0, -1)
    elif any(frmt in path for frmt in MATPLOTLIB_EXTENSIONS):
        data = plt.imread(path)
        metadata = None
        bounds = None
    elif any(frmt in path for frmt in NETCDF4_EXTENSIONS):
        data = netCDF4.Dataset(path, 'r')
        metadata = None
        bounds = None
    elif any(frmt in path for frmt in NUMPY_EXTENSIONS):
        data = np.load(path)
        metadata = None
        bounds = None
    else:
        data = None
        metadata = None
        bounds = None
        raise Exception('Error: file can not be opened or format not supported!')
        
    return data, metadata, bounds



import numpy as np


def get_patches(img : np.ndarray, kernel : tuple = (16, 16), stride : tuple = (16, 16)) -> np.ndarray:
    '''
        Split an image into patches

        Parameters:
        -----------
            - img : np.ndarray 
                a WxHxB image, with width W, height H and B bands
            - kernel : tuple
                Tuple of two values used to define the size of the patch
            - stride :  tuple
                Tuple of two values representing the stride to extract patches

        Returns:
        --------
            - patches : np.ndarray
                the Nx(kernel[0])x(kernel[1])xB vector containing the N patches extracted from img
        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
            [
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
            [
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        patches = get_patches(img, kernel=(2, 2), stride=(1,1))
        ```

        Output:
        -------
        ```python
        [[[[0.1 0.1 0.1]
        [0.2 0.2 0.2]]

        [[0.4 0.4 0.4]
        [0.5 0.5 0.5]]]


        [[[0.2 0.2 0.2]
        [0.3 0.3 0.3]]

        [[0.5 0.5 0.5]
        [0.6 0.6 0.6]]]


        [[[0.3 0.3 0.3]
        [0.3 0.3 0.3]]

        [[0.6 0.6 0.6]
        [0.6 0.6 0.6]]]


        [[[0.4 0.4 0.4]
        [0.5 0.5 0.5]]

        [[0.7 0.7 0.7]
        [0.8 0.8 0.8]]]


        [[[0.5 0.5 0.5]
        [0.6 0.6 0.6]]

        [[0.8 0.8 0.8]
        [0.9 0.9 0.9]]]


        [[[0.6 0.6 0.6]
        [0.6 0.6 0.6]]

        [[0.9 0.9 0.9]
        [0.9 0.9 0.9]]]


        [[[0.7 0.7 0.7]
        [0.8 0.8 0.8]]

        [[0.7 0.7 0.7]
        [0.8 0.8 0.8]]]


        [[[0.8 0.8 0.8]
        [0.9 0.9 0.9]]

        [[0.8 0.8 0.8]
        [0.9 0.9 0.9]]]


        [[[0.9 0.9 0.9]
        [0.9 0.9 0.9]]

        [[0.9 0.9 0.9]
        [0.9 0.9 0.9]]]]
        ```
    '''

    if len(img.shape) != 3:
        raise Exception("Error: len of img.shape must be 3")
    if kernel[0] < 1 or kernel[1] < 1:
        raise Exception("Error: kernel must be grather than 1")
    if stride[0] < 1 or stride[1] < 1:
        raise Exception("Error: kernel must be grather than 1")
    

    w, h, c = img.shape


    if kernel[0] > w or kernel[1] > h:
        raise Exception("Error: kernel size cannot be grather than image size")
    if stride[0] > w or stride[1] > h:
        raise Exception("Error: stride cannot be grather than image size")
    

    
    N = len(range(0, w, stride[0])) * len(range(0, h, stride[1]))
    patches = np.zeros((N, kernel[0], kernel[1], c))

    ct = 0
    for ww in range(0, w, stride[0]):
        for hh in range(0,  h, stride[1]):         
            patches[ct, ...] = img[ww:ww+kernel[0],hh:hh+kernel[1],:]
            ct += 1

    return patches


def patches_generator(img : np.ndarray, kernel : tuple = (16, 16), stride : tuple = (16, 16), batch_size : int = 1) -> np.ndarray:
    '''
        Split an image into patches

        Parameters:
        -----------
            - img : np.ndarray 
                a WxHxB image, with width W, height H and B bands
            - kernel : tuple
                Tuple of two values used to define the size of the patch
            - stride :  tuple
                Tuple of two values representing the stride to extract patches
            - batch_size : int
                integere representing how many patches yield at each iteration

        Returns:
        --------
            - patches : np.ndarray
                the (batch_size)x(kernel[0])x(kernel[1])xB vector containing the N patches extracted from img
        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
            [
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
            [
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        patches = patches_generator(img, kernel=(2, 2), stride=(1,1), batch_size = 1)
        ```

        Output:
        -------
        ```python
        
        ```
    '''

    if len(img.shape) != 3:
        raise Exception("Error: len of img.shape must be 3")
    if kernel[0] < 1 or kernel[1] < 1:
        raise Exception("Error: kernel must be grather than 1")
    if stride[0] < 1 or stride[1] < 1:
        raise Exception("Error: kernel must be grather than 1")
    if batch_size < 1:
        raise Exception("Error: batch_size must be grather than 1")
    

    w, h, c = img.shape


    if kernel[0] > w or kernel[1] > h:
        raise Exception("Error: kernel size cannot be grather than image size")
    if stride[0] > w or stride[1] > h:
        raise Exception("Error: stride cannot be grather than image size")
    
    
    patches = np.zeros((batch_size, kernel[0], kernel[1], c))   

    
    ct = 0 
    for ww in range(0, w-stride[0], stride[0]):
        for hh in range(0,  h-stride[1], stride[1]):

            patches[ct, :kernel[0], :kernel[1], :] = img[ww:ww+kernel[0],hh:hh+kernel[1],:]

            if ct == batch_size-1:
                ct = 0
                yield patches
            else: ct += 1


from tkinter.filedialog import askopenfilename
from tkinter import Tk


def get_path_gui() -> str:
    
    '''
        Get the path of a file using a graphical menu    
        
        Parameters:
        -----------
        Nothing

        Returns:
        --------
            - path: path of the image

        Usage:
        ------

        ```python
        path = get_path_gui()
        ```

        Output:
        -------
        ```
        /path/to/file/img.png
        ```

    '''
    
    root = Tk()
    root.withdraw()
    path = askopenfilename(
        title='Select the file or the folder!',
        initialdir='/'
        )
    root.destroy()
    
    return path


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 17:50:18 2022

@author: alessandrosebastianelli
"""

import numpy as np

def fft2d(image):
    '''
    
        Calculates the 2D Fast Fourier Transform of an image
        
        Input:
            - image: a WxH image, must be a single band image, with width W, height H
        Outputs:
            - fft: 2D spectrum of image
    '''
    
    fft = np.fft.fftshift(np.fft.fft2(image))
    
    return fft

def ifft2d(fft):
    '''
    
        Calculates the inverse 2D Fast Fourier Transform of a spectrum of an image
        
        Input:
            - fft: 2D spectrum of image 
        Outputs:
            - ifft: a WxH image
    '''
    
    ifft = np.abs(np.fft.ifft2(np.fft.ifftshift(fft)))
    
    return ifft
        
def fft3d(image):
    '''
    
        Calculates the 2D Fast Fourier Transform of an image with multiple bands
        
        Input:
            - image: a WxHxB image, with width W, height H and B bands
        Outputs:
            - fft: array of 2D spectra of the image
    '''
    
        
    ffts = []    
    for b in range(image.shape[-1]):
        ffts.append(fft2d(image[:,:,b]))
        
    ffts = np.array(ffts)
    ffts = np.moveaxis(ffts, 0, -1)
        
    return ffts 

def ifft3d(ffts):
    '''
    
        Calculates the inverse 2D Fast Fourier Transform of a vecctor of spectra of a multibands image
        
        Input:
            - fft: array of 2D spectra of the image
        Outputs:
            - ifft: a WxHxB image, with width W, height H and B bands
    '''
        
    iffts = []    
    for b in range(ffts.shape[-1]):
        iffts.append(ifft2d(ffts[:,:,b]))
        
    iffts = np.array(iffts)
    iffts = np.moveaxis(iffts, 0, -1)
        
    return iffts


import numpy as np


def percentile_prescaler(img : np.ndarray , perc : list, mmin : list = None) -> np.ndarray:
    '''
        Clip image between minimum and maximum based on the percentile
    
        Parameters:
        -----------
            - img : np.ndarray 
                a WxHxB image, with W width, H height and B bands            
            - perc: list
                the percentile values (one for each band)
            - mmin : list
                the minimum values (one for each band) to clip the img (default : None), if None mmin is calculated from img
        
        Returns:
        --------
            - img : np.ndarray 
                normalized WxHxB image, with W width, H height and B bands

        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
                [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        img_n = percentile_prescaler(img, perc=[95,95], mmin=[0,0])
        ```

        Output:
        -------
        ```
            [[0.1  1.1 ] 
            [0.2  1.2 ]  
            [0.3  1.3 ]]  
            [[0.4  1.4 ]  
            [0.5  1.5 ]  
            [0.6  1.6 ]]  
            [[0.7  1.7 ]  
            [0.8  1.8 ]  
            [0.86 1.86]]]  
        ```
    '''

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')
    
    if mmin is not None:
        if len(mmin) != img.shape[-1]:
            raise Exception('Error: lenght of mmin must be equals to number of bands')
    
    if len(perc) != img.shape[-1]:
        raise Exception('Error: lenght of perc must be equals to number of bands')


    if mmin is None: mmin = np.nanmin(img, axis=-1)

    for i in range(img.shape[-1]):
        mmax = np.nanpercentile(img[:,:,i], perc[i])
        img[:,:,i] = np.clip(img[:,:,i], mmin[i], mmax)
    
    return img


def minmax_scaler(img : np.ndarray, mmin : list = None, mmax : list = None, clip : list = [None, None]) -> np.ndarray: 
    '''
        Apply the min max scaler to the input img:  
        
        out = (img - minimum)/(maximum - minimum + E)          (1)  
        
        where E stabilizes the division. 
        
        Parameters:
        -----------
             - img : np.ndarray 
                a WxHxB image, with W width, H height and B bands
            - mmin : list 
                the minimum in equation (1) (one for each band) (default : None)
            - mmax : lsit
                the maximum in equation (1) (one for each band) (default : None)
            - clip : list 
                a list of two values used to constrain the image values (default : [None, None])
        
        Returns:
        --------
            - img : np.ndarray
                normalized WxHxB image, with W width, H height and B bands

        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
                [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        img_n = minmax_scaler(img, mmin=[0,0], mmax=[1,1], clip = [0,0.5])
        ```

        Output:
        -------
        ```
        [[[0.0999001 0.5      ]  
        [0.1998002 0.5      ]  
        [0.2997003 0.5      ]]  
        [[0.3996004 0.5      ]  
        [0.4995005 0.5      ]  
        [0.5       0.5      ]]  
        [[0.5       0.5      ]  
        [0.5       0.5      ]  
        [0.5       0.5      ]]]  
        ```
    '''

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')
    
    if mmin is not None:
        if len(mmin) != img.shape[-1]:
            raise Exception('Error: lenght of mmin must be equals to number of bands')
    
    if mmax is not None:
        if len(mmax) != img.shape[-1]:
            raise Exception('Error: lenght of mmax must be equals to number of bands')
    
    E = 0.001

    if mmin == None: mmin = np.nanmin(img, axis=-1)
    if mmax == None: mmax = np.nanmax(img, axis=-1)
    
    for i in range(img.shape[-1]):    
        img[:,:,i] = (img[:,:,i] - mmin[i])/((mmax[i] - mmin[i]) + E)
    
    if clip is not None: img = np.clip(img, clip[0], clip[-1])
    
    return img


def std_scaler(img : np.ndarray, mmean : list = None, sstd : list = None, clip : list = [None, None]) -> np.ndarray:
    '''
        Apply the min max scaler to the input img:  
        
        out = (img - mean)/(std)             (1)  
         
        Parameters:
        -----------
             - img : np.ndarray 
                a WxHxB image, with W width, H height and B bands
            - mmean : list 
                the mean in equation (1) (one for each band) (default : None)
            - sstd : lsit
                the standar deviation in equation (1) (one for each band) (default : None)
            - clip : list 
                a list of two values used to constrain the image values (default : [None, None])
        
        Returns:
        --------
            - img : np.ndarray
                normalized WxHxB image, with W width, H height and B bands

        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
                [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        img_n = std_scaler(img, None, None, clip = [0,0.5])
        ```

        Output:
        -------
        ```
            [[[0.  0.4]  
            [0.  0.4]  
            [0.  0.4]]  
            [[0.  0.5]  
            [0.  0.5]  
            [0.  0.5]]  
            [[0.2 0.5]  
            [0.2 0.5]  
            [0.2 0.5]]]   
        ```
    '''

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')
    
    if mmean is not None:
        if len(mmean) != img.shape[-1]:
            raise Exception('Error: lenght of mmin must be equals to number of bands')
    
    if sstd is not None:
        if len(sstd) != img.shape[-1]:
            raise Exception('Error: lenght of mmax must be equals to number of bands')
    
    if mmean == None: mmean = np.nanmean(img, axis=-1)
    if sstd  == None: sstd  = np.nanstd(img, axis=-1)
    
    for i in range(img.shape[-1]):    
        img[:,:,i] = (img[:,:,i] - mmean[i])/(sstd[i])
    
    if clip is not None: img = np.clip(img, clip[0], clip[-1])
    
    return img



import numpy as np


def percentile_prescaler(img : np.ndarray , perc : int, mmin : float = None) -> np.ndarray:
    '''
        Clip image between minimum and maximum based on the percentile
    
        Parameters:
        -----------
            - img : np.ndarray 
                a WxHxB image, with W width, H height and B bands            
            - perc: int
                the percentile value 
            - mmin : float
                the minimum value to clip the img (default : None), if None mmin is calculated from img
        
        Returns:
        --------
            - img : np.ndarray 
            normalized WxHxB image, with W width, H height and B bands

        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
                [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        img_n = percentile_prescaler(img, perc=95, mmin=0)
        ```

        Output:
        -------
        ```
            [[[0.1   1.1  ]  
            [0.2   1.2  ]  
            [0.3   1.3  ]]  
            [[0.4   1.4  ]  
            [0.5   1.5  ]  
            [0.6   1.6  ]]  
            [[0.7   1.7  ]  
            [0.8   1.8  ]  
            [0.9   1.815]]]  
        ```
    '''

    
    if mmin is None: mmin = np.nanmin(img)
    mmax = np.nanpercentile(img, perc)
    img = np.clip(img, mmin, mmax)
    
    return img


def minmax_scaler(img : np.ndarray, mmin : float = None, mmax : float = None, clip : list = [None, None]) -> np.ndarray: 
    '''
        Apply the min max scaler to the input img:  
        
        out = (img - minimum)/(maximum - minimum + E)          (1)  
        
        where E stabilizes the division. 
        
        Parameters:
        -----------
             - img : np.ndarray 
                a WxHxB image, with W width, H height and B bands
            - mmin : list 
                the minimum in equation (1) (default : None)
            - mmax : lsit
                the maximum in equation (1) (default : None)
            - clip : list 
                a list of two values used to constrain the image values (default : [None, None])
        
        Returns:
        --------
            - img : np.ndarray
                normalized WxHxB image, with W width, H height and B bands

        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
                [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        img_n = minmax_scaler(img, mmin=0, mmax=1, clip = [0,0.5])
        ```

        Output:
        -------
        ```
        [[[0.0999001 0.5      ]  
        [0.1998002 0.5      ]  
        [0.2997003 0.5      ]]  
        [[0.3996004 0.5      ]  
        [0.4995005 0.5      ]  
        [0.5       0.5      ]]  
        [[0.5       0.5      ]  
        [0.5       0.5      ]  
        [0.5       0.5      ]]]  
        ```
    '''
    
    E = 0.001
    if mmin == None: mmin = np.nanmin(img)
    if mmax == None: mmax = np.nanmax(img)
    img = (img - mmin)/((mmax - mmin) + E)
    if clip is not None: img = np.clip(img, clip[0], clip[-1])
    
    return img


def std_scaler(img : np.ndarray, mmean : float = None, sstd : float = None, clip : list = [None, None]) -> np.ndarray:
    '''
        Apply the min max scaler to the input img:  
        
        out = (img - mean)/(std)             (1)  
         
        Parameters:
        -----------
             - img : np.ndarray 
                a WxHxB image, with W width, H height and B bands
            - mmean : list 
                the mean in equation (1) (default : None)
            - sstd : lsit
                the standar deviation in equation (1) (default : None)
            - clip : list 
                a list of two values used to constrain the image values (default : [None, None])
        
        Returns:
        --------
            - img : np.ndarray
                normalized WxHxB image, with W width, H height and B bands

        Usage:
        ------
        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
                ],
                [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
                ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        img_n = std_scaler(img, None, None, clip = [0,0.5])
        ```

        Output:
        -------
        ```
            [[[0.         0.17770466]  
            [0.         0.35540933]  
            [0.         0.5       ]]  
            [[0.         0.5       ]  
            [0.         0.5       ]  
            [0.         0.5       ]]  
            [[0.         0.5       ]  
            [0.         0.5       ]  
            [0.         0.5       ]]]   
        ```
    '''
    
    if mmean == None: mmean = np.nanmean(img)
    if sstd  == None: sstd  = np.nanstd(img)
    img = (img - mmean)/(sstd)
    
    if clip is not None: img = np.clip(img, clip[0], clip[-1])
    
    return img



import numpy as np

def gaussian_filter(fft : np.ndarray, mx : int =0, my : int = 0, sx : int = 1, sy : int = 1, invert : bool = False) -> np.ndarray:
    '''
        Apply a 2D gaussian filter to the input 2D spectrum
        
        Prameters:
        ---------
            - fft : np.ndarray 
                2D spectrum to be filtered
            - mx : int 
                mean of gaussian x function (default : 0)
            - my : int
                mean of gaussian y function (default : 0)
            - sx : int
                standard deviation of gaussian x function (default : 1)
            = sy : int
                standard deviation of gaussian y function (default : 1)
            - invert : bool
                invert the distribution (default : False)
        
        Returns:
        --------
            - filt : np.ndarray
                filtered spectrum
        
        Usage:
        ------
        ```python
        import numpy as np

        fft = np.array(
            [
            [0,   0, 1,   0,0],
            [0, 0.5, 1, 0.5,0],
            [1,   1, 1,   1,1],
            [0, 0.5, 1, 0.5,0],
            [0,   0, 1,   0,0],
            ]

        )

        filt = gaussian_filter(fft, mx=0, my=0, sx=1, sy=1, invert=False)
        ```

        Output:
        -------
        ```
        array([  
            [0.        , 0.        , 0.09653235, 0.        , 0.        ],  
            [0.        , 0.061975  , 0.14045374, 0.061975  , 0.        ],  
            [0.09653235, 0.14045374, 0.15915494, 0.14045374, 0.09653235],  
            [0.        , 0.061975  , 0.14045374, 0.061975  , 0.        ],  
            [0.        , 0.        , 0.09653235, 0.        , 0.        ]])  
        ```
    '''

    if len(fft.shape) < 2:
        raise Exception("Error: the shape of fft must be greater than 2")
    
    x = np.linspace(-1, 1, fft.shape[0])
    y = np.linspace(-1, 1, fft.shape[1])
    
    x, y = np.meshgrid(x,y)

    fxy = 1. / (2. * np.pi * sx * sy) * np.exp(-((x - mx)**2. / (2. * sx**2.) + (y - my)**2. / (2. * sy**2.)))

    if invert: fxy = 1-fxy
    
    filt = fft*fxy
    
    return filt


def lhp_filter(fft, radius=0.5, invert=False):
    '''
        Apply a low pass or high pass filter to the input 2D spectrum
        
        Parameters:
        -----------
            - fft : np.ndarray
                2D spectrum to be filtered
            - radius : int
                size of the filter, ammited range [0-1], will mask to zero fft values lower than radius (default : 0.5)
            - invert: bool
                invert the distribution (default : False)

        Returns:
        --------
            - fft : np.ndarray
                filtered spectrum

        
        Usage:
        ------
        ```python
        import numpy as np

        fft = np.array(
            [
            [0,   0, 1,   0,0],
            [0, 0.5, 1, 0.5,0],
            [1,   1, 1,   1,1],
            [0, 0.5, 1, 0.5,0],
            [0,   0, 1,   0,0],
            ]

        )

        filt = lhp_filter(fft, radius=0.8, invert=False)
        ```

        Output:
        -------
        ```
        array([  
            [0., 0., 1., 0., 0.],  
            [0., 0., 0., 0., 0.],  
            [1., 0., 0., 0., 1.],  
            [0., 0., 0., 0., 0.],  
            [0., 0., 1., 0., 0.]])  
        ```
    '''
    
    if len(fft.shape) < 2:
        raise Exception("Error: the shape of fft must be greater than 2")
    
    x = np.linspace(-1, 1, fft.shape[0])
    y = np.linspace(-1, 1, fft.shape[1])
    x, y = np.meshgrid(x, y)

    fxy = np.sqrt(x**2 + y**2)

    if invert: fxy = 1-fxy

    fft[fxy<radius] = 0
    
    return fft


import numpy as np

def geo_plot(img : np.ndarray, metadata : dict, bounds : list) -> None:
    '''
        Overlay a satellite image on a map

        Parameters:
        -----------
            - img : np.ndarray
                the WxHxC image to be plotted, with W width, H height and B bands (channel last)
            - metadata : dict
                metadata for the image to be saved
            - bounds : list
                list of geo bounds for the image
        
        Returns:
        --------
        Nothing an image is displayed

        Usage:
        ------
        ```python
        import numpy as np

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
             ]
            ]  
        )  
            
        meta = {'driver': 'GTiff',  
                'dtype': 'float64',  
                'nodata': None,  
                'width': 3,  
                'height': 3,  
                'count': 2,  
                'crs': CRS.from_epsg(32632),  
                'transform': Affine(10.0, 0.0, 638640.0, 0.0, -10.0, 5084590.0),  
                'blockxsize': 3,  
                'blockysize': 3,  
                'tiled': True,  
                'compress': 'lzw',  'interleave': 'pixel'  
                }  

        bounds = BoundingBox(left=638640.0, bottom=5074190.0, right=649070.0, top=5084590.0))  

        geo_plot(img, meta, bounds)
        ```

        Output:
        ------
        Nothing, the image is displayed
    '''
    pass


import numpy as np

def normalized_difference(channel_1 : np.ndarray, channel_2 : np.ndarray) -> np.ndarray:
    '''
        Normalized Difference
        
        nd = (channelA - channleB)/(channelA + channle_B + E)          (1)
        
        where E stabilizes the division.
        
        Parameters:
        -----------
            - channel_1 : np.ndarray 
                channelA in equation (1)
            - channel_2 : np.ndarray
                channelB in equation (1)
        
        Returns:
        --------
            - nd : np.ndarray
                normalized difference between channelA and Channelb in equation (1)            
        
        Usage:
        ------
        ```python
        import numpy as np

        ch1 = np.array(
                    [
                        [0.1, 0.2, 0.3],
                        [0.4, 0.5, 0.6],
                        [0.7, 0.8, 0.9]
                    ]
            )

        ch2 = np.array(
                    [
                        [2.1, 2.2, 2.3],
                        [2.4, 2.5, 2.6],
                        [2.7, 2.8, 2.9]
                    ]
            )

        nd = normalized_difference(ch1, ch2)
        ```

        Output:
        -------
        ```
            [[-0.90909091 -0.83333333 -0.76923077]  
            [-0.71428571 -0.66666667 -0.625     ]  
            [-0.58823529 -0.55555556 -0.52631579]]  
        ```

    '''
    
    num = channel_1 - channel_2
    den = channel_1 + channel_2
    
    # Avoiding 0 divisions
    den[den==0] = 0.001 #E
    
    nd = num/den
    
    return nd


import numpy as np

def dict_disp(dictionary : dict) -> None:
    '''
        Print a dictionary in a more readable way

        Parameter:
        ----------
            - dictionary : dict
                python dictionary to print
        
        Returns:
        --------
        Nothing, it will print the dictionary

        Usage:
        ------

        ```python
        dictionary = {'Name' : 'John', 'Surname' : 'Wick'}
        disp_dict(dictionary)
        ```

        Output:
        -------

        ```
        -----------Dictionary------------  
            [*] Name : Johnny
            [*] Surname : Wick
        ```

    '''

    l_k = len([k for k in dictionary.keys() if len(str(k))==max([len(str(n)) for n in dictionary.keys()])][0])
    l_v = len([k for k in dictionary.values() if len(str(k))==max([len(str(n)) for n in dictionary.values()])][0])

    n =  20 + l_k + l_v
    print('{:-^{len}}'.format('Dictionary', len=n))

    for key, value in dictionary.items():
        print('\t [*] {} : {}'.format(str(key), str(value)))


def print_stats(img : np.ndarray, bandwise : bool = False) -> None:
    '''
        Print some statistics for the input image

        Parameters:
        -----------
            - image: a WxHxB image, with width W, height H and B bands
            - bandwise : bool 
                if true the function will print the statistics for each band separately (default : False)
        
        Returns:
        --------
        Nothing, it will print image statistics

        Usage:
        ------
        ```python
        import numpy as np

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
             ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        print_stats(img, bandwise = True)

        ```

        Output:
        -------

        ```
        ```
    '''

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 - (space, space, channels)')

    print('{:-^{len}}'.format('Image Statistics', len=50))
    print('{:<20s}{}'.format('[*] Shape', img.shape))

    if bandwise:
        for i in range(img.shape[-1]):
            print('{:<20s}'.format('Band '+str(i)))
            print('\t{:<20s}{:<30s}'.format('[*] Max',  str(np.max(img[:, :, i]))))
            print('\t{:<20s}{:<30s}'.format('[*] Min',  str(np.min(img[:, :, i]))))
            print('\t{:<20s}{:<30s}'.format('[*] Mean', str(np.mean(img[:, :, i]))))
            print('\t{:<20s}{:<30s}'.format('[*] Std',  str(np.std(img[:, :, i]))))
            print('\t{:<20s}{:<30s}'.format('[*] Med',  str(np.median(img[:, :, i]))))
    else:
        print('{:<20s}{:<30s}'.format('[*] Max',  str(np.max(img))))
        print('{:<20s}{:<30s}'.format('[*] Min',  str(np.min(img))))
        print('{:<20s}{:<30s}'.format('[*] Mean', str(np.mean(img))))
        print('{:<20s}{:<30s}'.format('[*] Std',  str(np.std(img))))
        print('{:<20s}{:<30s}'.format('[*] Med',  str(np.median(img))))


from sklearn.decomposition import PCA
import numpy as np

def image_PCA(img : np.ndarray, n_components : int) -> np.ndarray:
    '''

        Reduce the image dimensionality along the channells axis using PCA

        Parameters:  
        ----------  
            - img : np.ndarray
                a WxHxB image, with width W, height H and B bands (channel last)
            - n_components : int
                number of bands of the compressed image
        
        Returns:
        --------
            - reduced : np.ndarray
                the WxHxn_components compressed image, with width W, height H and n_components bands

        Usage:
        ------

        ```python
        import numpy as np  

        img         = np.array(  
            [[
                [0.1, 0.2, 0.3],  
                [0.4, 0.5, 0.6],  
                [0.7, 0.8, 0.9]
             ],
             [
                [1.1, 1.2, 1.3],  
                [1.4, 1.5, 1.6],  
                [1.7, 1.8, 1.9]
             ]
            ]  
        ) 

        # Making channels last
        img = np.moveaxis(img, 0, -1)

        reduced = image_PCA(img, n_components = 1)
        ```
        Output:
        ------
        ```
        array([
              [[ 0.56568542],
                [ 0.42426407],
                [ 0.28284271]],
              [[0.14142136],
                [-0.        ],
                [-0.14142136]],
              [[-0.28284271],
                [-0.42426407],
                [-0.56568542]]])
       ```

    '''

    if len(img.shape) != 3:
        raise Exception('Error: lenght of image shape must be 3 -  (space, space, channels)')

    pca = PCA(n_components)
    reduced = pca.fit_transform(img.reshape((img.shape[0] * img.shape[1], img.shape[2])))
    reduced = reduced.reshape((img.shape[0], img.shape[1], n_components))
    reduced = np.array(reduced)

    return reduced


def image_stack_PCA(imgs : np.ndarray, n_components : int) -> np.ndarray:
    '''
        Reduce a temporal stack of images along the temporal axis
    '''
    if len(imgs.shape) != 4:
        raise Exception('Error: lenght of imgs shape must be 4 - (space, space, channels, time)')
    
    pass